/* $Id: common-init.cpp 451 2011-02-26 10:32:30Z oh2gve $
 * 
 * Copyright 2007, 2008, 2009, 2010, 2011 Tapio Aaltonen
 * 
 *
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


// Build config.
#ifdef __WXMSW__
#	include "windows/config.h"
#endif
#ifdef HAVE_CONFIG_H
#		include <config.h>
#	endif

// Own stuff.
#include "main.h"
#include "kisstnc.h"
#include "uplink.h"
#ifndef __WXMSW__
#	include "linux/ax25port.h"
#endif

// Std stuff.
#include <string>
using std::string;
#include <fstream>
using std::ofstream;
#include <ios>
using std::ios;
#include <sstream>
using std::ostringstream;
#include <vector>
using std::vector;
#ifndef __WXMSW__
#	include <csignal>
#endif

// Wx stuff.
#include <wx/log.h>



// LOG HANDLING

bool AprsApp::openLog(bool const initialOpening)
{  
	// Open log file.
	logFile_ = new ofstream(configuration_.logcfg.mainFileName.ToAscii(), ios::app);

	// Check that it opened. Fail if not.
	if ( !logFile_ || !(*logFile_) )
	{  
		wxLogError(wxT("Error: failed to open log file ") + configuration_.logcfg.mainFileName + wxT("."));
		delete logFile_;
		logFile_ = NULL;
		return false;
	}

	// Construct a logging target ant set a nice timestamp format.
	wxLog* wxLogObjectPtr = new wxLogStream(logFile_);
	if ( !wxLogObjectPtr )
	{
		wxLogError(wxT("Error: failed to initialize log file handling."));
		delete logFile_;
		logFile_ = NULL;
		return false;
	}
	wxLogObjectPtr->SetTimestamp(wxT("%Y-%m-%d %H:%M:%S"));
	
	// Set our log as default logging target. This returns pointer to previous
	// target, which must be deleted.
	delete wxLog::SetActiveTarget(wxLogObjectPtr);
	wxLogMessage(wxT("Log file opened."));

#ifndef __WXMSW__
	// If this is first time the log is opened, close stdout and stderr.
	if ( initialOpening )
	{
		if ( dup2(nullFd_, 0) < 0 || dup2(nullFd_, 1) < 0 || dup2(nullFd_, 2) < 0 )
		{  
			// This has never happened.
			wxLogError(wxT("Error: Closing stdio failed."));
			return false;
		}
	}
#endif
	
	return true;
}


void AprsApp::closeLog()
{
	if ( logFile_ != NULL )
	{
		// Remove log file from log target and close it.
		wxLogMessage(wxT("Closing log file.\n"));
		delete wxLog::SetActiveTarget(NULL);   
		delete logFile_;
		logFile_ = NULL;
	}
}


// CONNECTION HANDLING


bool AprsApp::openConnections()
{
	// Construct our login line.
	string greet = "user ";
	greet += configuration_.callsign.ToAscii();
	greet += " pass ";
	greet += configuration_.uplink.passcode.ToAscii();
	greet += " vers ";
	greet += PACKAGE;
	greet += " ";
	greet += PACKAGE_VERSION;
	if ( !configuration_.uplink.filter.IsEmpty() )
	{
		greet += " filter ";
		greet += configuration_.uplink.filter.ToAscii();
	}
	greet += "\r\n";

	// First open the server connection.
	wxLogMessage(wxT("Connecting uplink."));

	// Create the thread object.
	uplink_ = new Uplink(configuration_.uplink, greet, packetlog_, dupefilter_);
	if ( uplink_->Create() != wxTHREAD_NO_ERROR )
	{  
		delete uplink_;
		uplink_ = 0;
		wxLogError(wxT("Error: cannot create thread for uplink."));
		return false;
	}
	
	// Start the thread.
	uplink_->Run();
	uplink_->waitForInit();
	
	// Check that it managed to open it's connection.
	if ( !(*uplink_) )
	{
		wxLogError(wxT("Error: uplink could not be opened."));
		
		// Stop the connection thread, and wait for it to die.
		uplink_->terminate();
		while ( uplink_->IsRunning() )
		{
			Yield();
		}
		delete uplink_;
		uplink_ = NULL;
		return false;
	}
	
	// Order beaconing to uplink.
	vector<wxString> path;
	path.push_back(wxT("TCPIP*"));
	if ( !beacon_->addConnection(uplink_, 'u', path) )
	{
		uplink_->terminate();
		while ( uplink_->IsRunning() )
		{
			Yield();
		}
		delete uplink_;
		uplink_ = NULL;
		return false;
	}
	
	// Uplink successfully opened.
	wxLogMessage(wxT("Uplink initialized."));
	

	// Open other connections.

	// Temporary pointer for the connection to be opened.
	Connection* tmpConnPtr = NULL;

	// Loop through all given connection configurations.
	for ( unsigned int i = 0; i < configuration_.connections.size(); i++ )
	{  
		// Tell what we're going to do.
		wxLogMessage(wxString::Format(wxT("Opening connection %d."), i+1));
		ostringstream conId;
		
		// Find out what type of connection we are going to open.
		switch ( configuration_.connections.at(i).type )
		{
			
		 case Cfg::AX25PORT:
			{
#ifdef HAVE_LIBAX25
				tmpConnPtr = new Ax25Iface(configuration_.connections.at(i), i+48+1, configuration_.callsign, packetlog_, dupefilter_);
#else
				wxLogError(wxT("Linux AX.25 interface support is not enabled in this build."));
				return false;
#endif
				break;       
			}
		 case Cfg::SERIALKISSTNC:
			{
				tmpConnPtr = new KissTnc(configuration_.connections.at(i), i+48+1, static_cast<char const*>(configuration_.callsign.ToAscii()), packetlog_, dupefilter_, configuration_.lockdir);
				break;
			}
		 default:
			{
				wxLogError(wxString::Format(wxT("Unknown connection type for connection %d."), i+1));
				return false;
				break;
			}
		}
		
		// Check that port thread got created.
		if ( tmpConnPtr->Create() != wxTHREAD_NO_ERROR )
		{  
			delete tmpConnPtr;
			wxLogError(wxString::Format(wxT("Can't create thread for connection %d"), i+1));
			return false;
		}
		
		// Start the thread.
		tmpConnPtr->Run();
		// Wait for it to start up.
		tmpConnPtr->waitForInit();
		
		// If we're interested, check wether thread managed to open connection.
		if ( configuration_.connections.at(i).dieOnError && !(*tmpConnPtr) )
		{
			wxLogError(wxString::Format(wxT("Connection %d could not be opened"), i+1));
		
			// Stop the the connection thread and wait for it to stop.
			tmpConnPtr->terminate();
			while ( tmpConnPtr->IsRunning() )
			{
				Yield();
			}
			delete tmpConnPtr;
			return false;
		}
		
		// Save the port.
		connections_.push_back(tmpConnPtr);
		
		// Order beaconing to the port.
		beacon_->addConnection(tmpConnPtr, connections_.size()+48, configuration_.connections.at(i).path);

		// Tell that we were succesfull.
		wxLogMessage(wxString::Format(wxT("Connection %d opened."), i+1));
	}
	
	return true;
}


void AprsApp::closeConnections()
{
	// Close connections.
	for ( unsigned int i = 0; i < connections_.size(); i++ )
	{
		// Close the connection and wait for it to go.
		connections_.at(i)->terminate();
		while ( connections_.at(i)->IsRunning() )
		{
			Yield();
		}
		delete connections_.at(i);
		wxLogMessage(wxString::Format(wxT("Connection %d closed."), i+1));
	}
	
	// Close uplink.
	if ( uplink_ != 0 )
	{
		uplink_->terminate();
		while ( uplink_->IsRunning() )
		{
			Yield();
		}
		delete uplink_;
		wxLogMessage(wxT("Uplink closed."));
	}
}


// ASSISTANT THREADS


bool AprsApp::startAssistants()
{
	// Start packet logging.
	packetlog_ = new Packetlog(configuration_.logcfg, configuration_.callsign);
	if ( packetlog_->Create() != wxTHREAD_NO_ERROR )
	{
		wxLogError(wxT("Failed to create thread for packet logging."));
		return false;
	}
	packetlog_->Run();
	
        // Check if it started.
	packetlog_->waitForInit();
        if ( !(*packetlog_) )
        {
	        wxLogError(wxT("Failed to start packet logging."));
	        while ( packetlog_->IsRunning() )
	        {
	        	Yield();
		}
		delete packetlog_;
		packetlog_ = NULL;
	        return false;
        }
        
        // Start dupe filter.
        dupefilter_ = new DupeFilter(configuration_.duplicateTime);
        if ( dupefilter_->Create() != wxTHREAD_NO_ERROR )
        {
        	wxLogError(wxT("Failed to create thread for duplicate filtering."));
        	packetlog_->terminate();
        	while ( packetlog_->IsRunning() )
        	{
        		Yield();
		}
		delete packetlog_;
		packetlog_ = NULL;
		
		delete dupefilter_;
		dupefilter_ = NULL;
		
        	return false;
	}
	dupefilter_->Run();
	
	// Start beacon.
	beacon_ = new Beacon(configuration_, dupefilter_);
        if ( beacon_->Create() != wxTHREAD_NO_ERROR )
        {
        	wxLogError(wxT("Failed to create thread for beacon manager."));
        	while ( beacon_->IsRunning() )
        	{
        		Yield();
		}
		delete beacon_;
		beacon_ = NULL;
        	
        	packetlog_->terminate();
        	while ( packetlog_->IsRunning() )
        	{
        		Yield();
		}
		delete packetlog_;
		packetlog_ = NULL;

		dupefilter_->terminate();
		while ( dupefilter_->IsRunning() )
		{
			Yield();
		}
		delete dupefilter_;
		dupefilter_ = NULL;

        	return false;
	}
	beacon_->Run();
	
	// Check if it started.
	beacon_->waitForInit();
	if ( !(*beacon_) )
	{
		wxLogError(wxT("Failed to start beaconing."));
        	while ( beacon_->IsRunning() )
        	{
        		Yield();
		}
		delete beacon_;
		beacon_ = NULL;
		
        	packetlog_->terminate();
        	while ( packetlog_->IsRunning() )
        	{
        		Yield();
		}
		delete packetlog_;
		packetlog_ = NULL;

		dupefilter_->terminate();
		while ( dupefilter_->IsRunning() )
		{
			Yield();
		}
		delete dupefilter_;
		dupefilter_ = NULL;
		
        	return false;
	}
	
	// Assistant threads are go.
	return true;
}


void AprsApp::stopAssistants()
{
	if ( beacon_ != NULL )
	{
		beacon_->terminate();
		while ( beacon_->IsRunning() )
		{
			Yield();
		}
		delete beacon_;
		beacon_ = NULL;
	}

	if ( packetlog_ != NULL )
	{
		packetlog_->terminate();
		while ( packetlog_->IsRunning() )
		{
			Yield();
		}
		delete packetlog_;
		packetlog_ = NULL;
	}
	
	if ( dupefilter_ != NULL )
	{
		dupefilter_->terminate();
		while ( dupefilter_->IsRunning() )
		{
			Yield();
		}
		delete dupefilter_;
		dupefilter_ = NULL;
	}
}
