/* $Id: connection.h 453 2011-02-27 10:06:27Z oh2gve $
 *
 * Copyright 2007 Antti Seppänen
 * Copyright 2007, 2008, 2009, 2010, 2011 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


#ifndef CONNECTION_H
#define CONNECTION_H


// own stuff.
#include "cfg.h"
#include "packetlog.h"
#include "dupefilter.h"
#include "packet.h"
#include "aprspacket.h"

// std stuff.
#include <queue>

// wx stuff.
#include <wx/thread.h>


/// Common interface for connections.
/**
 * This class is responsible for all concurrency issues. It uses semaphores to
 * synchronize i/o and mutexes to protect it's critical sections.
 *
 * Functions in the public interface should be called only from the main
 * thread.
 */
class Connection : public wxThread
{
public:
	
	/// Constructor.
	Connection(Cfg::ConnectionState const& state, char const& conId, Packetlog* const& packetlog, DupeFilter* const& dupefilter);

	/// Destructor.
	virtual ~Connection();

	/// This is the entry point of thread. "Main()"-function.
	/**
	 * Must be implemented in child class.
	 */
	virtual ExitCode Entry() = 0;

	/// Blocks as long as initialization operations are underway.
	/**
	 * Don't call any of the other methods before this returns.
	 */
	void waitForInit();

	/// Get status of the connection.
	/**
	 * Returns true if thread has fatal error situation, false otherwise.
	*/
	bool operator!();
	
	/// Close connection and terminate thread.
	void terminate();

	/// Write a packet to the connection.
	/**
	 * \return True, if writing to output buffer succeeded. False, if not.
	 */
	void write(Packet* const& packet);

	/// Read a packet from the connection.
	/**
	 * \return A packet if there is one, NULL otherwise.
	 */
	Packet* read();

protected:

	/// Change status of this thread to started.
	/**
	 * Locks a mutex to protect startup of this thread. Call this on first
	 * line of Entry(). No other protected method may be called before this.
	 */
	void threadStarted();
	
	/// Relase mutex protecting thread's startup.
	/**
	 * Child classes must call this after trying to start up.
	 */
	void startupDone();
	
	/// Change status of this connection to error.
	void weHasError();
	
	/// Tells thread to terminate.
	/**
	 * Poll this and quit if you get false.
	*/
	bool keepGoing();
	
	/// Put data received from connection to input queue.
	void inputPush(AprsPacket* const& data);

	/// Check output queue for data to be transmitted.
	Packet* outputPop();

private:

	/// State of this connection.
	Cfg::ConnectionState const state_;
	
	/// ID of this connection.
	char const conId_;
	
	/// Packet logger.
	Packetlog* packetlog_;
	
	/// Dupe detector.
	DupeFilter* dupefilter_;
	
	/// Mutex for protecting private data of this connection.
	/**
	 * This is used to serialize operations on all private variables.
	 */
	wxMutex mutex_;
	
	/// False as long as we're not running yet.
	bool isStarted_;
	
	/// Error flag of this connection.
	bool hasFatalError_;
	
	/// Status flag.
	bool keepgoing_;

	/// Buffer for incoming data.
	std::queue<Packet*> inputQueue_;
	/// Buffer for outgoing data.
	std::queue<Packet*> outputQueue_;
};

#endif // CONNECTION_H
