/* $Id: filter.cpp 443 2011-02-20 15:00:16Z oh2gve $
 * 
 * Copyright 2009, 2010, 2011 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

#ifdef HAVE_CONFIG_H
#	ifdef __WXMSW__
#		include "../windows/config.h"
#	else
#		include <config.h>
#	endif
#endif

// own stuff.
#include "filter.h"
#include "filterelement.h"
#include "aprspacket.h"
/*#include "packet.h"
*/
// std stuff.
#include <vector>
using std::vector;
#include <string>
using std::string;
#include <map>
using std::map;
/*#include <list>
using std::list;
using std::make_pair;
#include <ctime>
*/
// wx stuff.
#include <wx/log.h>



Filter::Filter(vector<FilterElement> const& elements):
elements_(elements), clients_()
{
}


Filter::~Filter()
{
}


void Filter::incomingPacket(AprsPacket const* packet)
{
	// Check if packet contains location.
	AprsPacket::Location loc;
	bool hasLocation = false;
	if ( packet->getLocation(loc) )
	{
		hasLocation = true;
	}
	
	// Choose treatment based on packet type.
	if ( hasLocation )
	{
		// Check packet against all filter elements.
		for ( unsigned int i = 0; i < elements_.size(); ++i )
		{
			if ( elements_.at(i).accepts(packet->getLabel(), loc.lon, loc.lat) )
			{
				// Add or update timestamp.
				updateClientList(packet->getLabel());
				break;
			}
		}
	}
	else
	{
		// Check if target is already on list.
		if ( onClientList(packet->getLabel()) )
		{
			// Update timestamp.
			updateClientList(packet->getLabel());
		}
		else
		{
			// Check packet againts only location-independent filters.
			for ( unsigned int i = 0; i < elements_.size(); ++i )
			{
				if ( elements_.at(i).accepts(packet->getLabel()) )
				{
					// Add to list.
					updateClientList(packet->getLabel());
					break;
				}
			}
		}
	}
}


bool Filter::outgoingPacket(AprsPacket const* packet)
{
	// Match the packet.
	return passesStaticFilter(packet) || passesDynamicFilter(packet);
}


// PRIVATE


bool Filter::passesStaticFilter(AprsPacket const* packet)
{
	// Check if packet contains location.
	AprsPacket::Location loc;
	bool hasLocation = false;
	if ( packet->getLocation(loc) )
	{
		hasLocation = true;
	}
	
	// Check packet against filter elements.
	for ( unsigned int i = 0; i < elements_.size(); ++i )
	{
		if ( hasLocation )
		{
			if ( elements_.at(i).accepts(packet->getLabel(), loc.lon, loc.lat) )
			{
				return true;
			}
		}
		else
		{
			if ( elements_.at(i).accepts(packet->getLabel()) )
			{
				return true;
			}
		}      
	}

	// There was no element to say yes.
	return false;
}


bool Filter::passesDynamicFilter(AprsPacket const* packet)
{
	// Flush old packages away from the dynamic filter.
	flush();
	
	// If our client sends packet only via other network, we put it to this network too.
	if ( onClientList(packet->getLabel()) )
	{
		return true;
	}
	
	// If packet is a message to our client, we pass it even if sender is not our client.
	string message, destination;
	if ( packet->getMessage(message, destination) )
	{
		// Check if destination is our client.
		if ( onClientList(destination) )
		{
			// Put message sender to our client list too, so that message receiver will get his position.
			updateClientList(packet->getLabel());
			
			return true;
		}
	}
	
	return false;
}



void Filter::updateClientList(string const& call)
{
	clients_[call] = time(NULL);
}


bool Filter::onClientList(string const& call) const
{
	map<string, time_t>::const_iterator pos = clients_.find(call);
	return pos != clients_.end();
}


void Filter::flush()
{
	// Flush client list.
	for ( map<string, time_t>::iterator i = clients_.begin(); i != clients_.end(); )
	{
		// Check if last info from the client is older than 30 mins.
		if ( time(NULL) - i->second > 60*30 )
		{
			clients_.erase(i++);
		}
		else
		{
			++i;
		}
	}
}
